@@ -14,8 +14,8 @@ Responsibility: Keith Stribley, Martin Hosken
 
     You should also have received a copy of the GNU Lesser General Public
     License along with this library in the file named "LICENSE".
-    If not, write to the Free Software Foundation, 51 Franklin Street, 
-    Suite 500, Boston, MA 02110-1335, USA or visit their web page on the 
+    If not, write to the Free Software Foundation, 51 Franklin Street,
+    Suite 500, Boston, MA 02110-1335, USA or visit their web page on the
     internet at http://www.fsf.org/licenses/lgpl.html.
 
 Alternatively, the contents of this file may be used under the terms of the
@@ -24,8 +24,8 @@ License, as published by the Free Software Foundation, either version 2
 of the License or (at your option) any later version.
 
 Description:
-A simple console app that creates a segment using FileFont and dumps a 
-diagnostic table of the resulting glyph vector to the console. 
+A simple console app that creates a segment using FileFont and dumps a
+diagnostic table of the resulting glyph vector to the console.
 If graphite has been built with -DTRACING then it will also produce a
 diagnostic log of the segment creation in grSegmentLog.txt
 -----------------------------------------------------------------------------*/
@@ -92,9 +92,10 @@ public:
     bool ws;
     bool rtl;
     bool useLineFill;
+    bool noprint;
     int useCodes;
+    bool autoCodes;
     int justification;
-    bool enableCache;
     float width;
     int textArgIndex;
     unsigned int * pText32;
@@ -105,7 +106,7 @@ public:
     char * alltrace;
     int codesize;
     gr_face_options opts;
-    
+
 private :  //defensive since log should not be copied
     Parameters(const Parameters&);
     Parameters& operator=(const Parameters&);
@@ -137,8 +138,9 @@ void Parameters::clear()
     ws = false;
     useLineFill = false;
     useCodes = 0;
+    autoCodes = false;
+    noprint = false;
     justification = 0;
-    enableCache = false;
     width = 100.0f;
     pText32 = NULL;
     textArgIndex = 0;
@@ -156,7 +158,7 @@ void Parameters::closeLog()
 {
   if (log==stdout)
     return ;
-  
+
   fclose(log);
   log = stdout;
 }
@@ -200,7 +202,7 @@ bool Parameters::loadFromArgs(int argc, char *argv[])
     codesize = 4;
     bool argError = false;
     char* pText = NULL;
-    typedef enum 
+    typedef enum
     {
         NONE,
         POINT_SIZE,
@@ -210,6 +212,7 @@ bool Parameters::loadFromArgs(int argc, char *argv[])
         LINE_END,
         LINE_FILL,
         CODES,
+        AUTOCODES,
         FEAT,
         LOG,
         TRACE,
@@ -342,11 +345,6 @@ bool Parameters::loadFromArgs(int argc, char *argv[])
                     option = NONE;
                     ws = true;
                 }
-                else if (strcmp(argv[a], "-cache") == 0)
-                {
-                    option = NONE;
-                    enableCache = true;
-                }
                 else if (strcmp(argv[a], "-feat") == 0)
                 {
                     option = FEAT;
@@ -355,6 +353,10 @@ bool Parameters::loadFromArgs(int argc, char *argv[])
                 {
                     option = SIZE;
                 }
+                else if (strcmp(argv[a], "-noprint") == 0)
+                {
+                    noprint = true;
+                }
                 else if (strcmp(argv[a], "-codes") == 0)
                 {
                     option = NONE;
@@ -364,6 +366,18 @@ bool Parameters::loadFromArgs(int argc, char *argv[])
                     pText32 = (unsigned int *)malloc(sizeof(unsigned int) * argc);
                     fprintf(log, "Text codes\n");
                 }
+                else if (strcmp(argv[a], "-auto") == 0)
+                {
+                    const unsigned kCodeLimit = 0x1000;
+                    charLength = kCodeLimit - 1;
+                    pText32 = (unsigned int *)malloc(sizeof(unsigned int) * kCodeLimit);
+                    unsigned int i;
+                    for (i = 1; i < kCodeLimit; ++i)
+                        pText32[i - 1] = i;
+                    pText32[charLength] = 0;
+                    autoCodes = true;
+                    option = NONE;
+                }
                 else if (strcmp(argv[a], "-linefill") == 0)
                 {
                     option = LINE_FILL;
@@ -409,6 +423,7 @@ bool Parameters::loadFromArgs(int argc, char *argv[])
                 if (code > 0)
                 {
 // convert text to utfOut using iconv because its easier to debug string placements
+                    assert(pText32);
                     pText32[charLength++] = code;
                     if (charLength % 10 == 0)
                         fprintf(log, "%4x\n",code);
@@ -437,12 +452,12 @@ bool Parameters::loadFromArgs(int argc, char *argv[])
     if (mainArgOffset < 1) argError = true;
     else if (mainArgOffset > 1)
     {
-        if (!useCodes && pText != NULL)
+        if (!autoCodes && !useCodes && pText != NULL)
         {
             charLength = convertUtf<gr2::utf8>(pText, pText32);
             if (!pText32)
             {
-            	if (charLength == -1)
+            	if (charLength == ~0)
             		perror("decoding utf-8 data failed");
             	else
             		perror("insufficent memory for text buffer");
@@ -462,8 +477,9 @@ bool Parameters::loadFromArgs(int argc, char *argv[])
             }
             fprintf(log, "\n");
         }
-        else 
+        else
         {
+            assert(pText32);
             pText32[charLength] = 0;
             fprintf(log, "\n");
         }
@@ -517,7 +533,7 @@ void Parameters::printFeatures(const gr_face * face) const
         fprintf(log, "\t");
         for (size_t j = 4; j; --j)
         {
-        	const char c = lang_id >> (j*8-8);
+        	const unsigned char c = lang_id >> (j*8-8);
             if ((c >= 0x20) && (c < 0x80))
                 fprintf(log, "%c", c);
         }
@@ -533,7 +549,6 @@ gr_feature_val * Parameters::parseFeatures(const gr_face * face) const
     if (features && (pLang = strstr(features, "lang=")))
     {
         pLang += 5;
-        size_t i = 0;
         for (int i = 4; i; --i)
         {
             lang_id <<= 8;
@@ -557,6 +572,7 @@ gr_feature_val * Parameters::parseFeatures(const gr_face * face) const
         {
             case ',':
             case '&':
+                assert(valueText);
                 value = atoi(valueText);
                 if (ref)
                 {
@@ -575,6 +591,7 @@ gr_feature_val * Parameters::parseFeatures(const gr_face * face) const
                 }
                 if (!ref)
                 {
+                    assert(name);
                     feat_id = atoi(name);
                     ref = gr_face_find_fref(face, feat_id);
                 }
@@ -591,6 +608,7 @@ gr_feature_val * Parameters::parseFeatures(const gr_face * face) const
         }
         if (ref)
         {
+            assert(valueText);
             value = atoi(valueText);
             gr_fref_set_feature_value(ref, value, featureList);
             if (feat_id > 0x20000000)
@@ -611,11 +629,17 @@ int Parameters::testFileFont() const
 //    try
     {
         gr_face *face = NULL;
-        if (alltrace) gr_start_logging(NULL, alltrace);
-        if (enableCache)
-            face = gr_make_file_face_with_seg_cache(fileName, 1000, opts | gr_face_dumbRendering);
+        FILE * testfile = fopen(fileName, "rb");
+        if (!testfile)
+        {
+            fprintf(stderr, "Unable to open font file\n");
+            return 4;
+        }
         else
-            face = gr_make_file_face(fileName, opts);
+            fclose(testfile);
+
+        if (alltrace) gr_start_logging(NULL, alltrace);
+        face = gr_make_file_face(fileName, opts);
 
         // use the -trace option to specify a file
     	if (trace)	gr_start_logging(face, trace);
@@ -641,10 +665,10 @@ int Parameters::testFileFont() const
             featureList = parseFeatures(face);
         if (codesize == 2)
         {
-            unsigned short *pText16 = (unsigned short *)malloc(textSrc.getLength() * 2 * sizeof(unsigned short));
+            unsigned short *pText16 = (unsigned short *)malloc((textSrc.getLength() * 2 + 1) * sizeof(unsigned short));
             gr2::utf16::iterator ui = pText16;
             unsigned int *p = pText32;
-            for (int i = 0; i < textSrc.getLength(); ++i)
+            for (unsigned int i = 0; i < textSrc.getLength(); ++i)
             {
                 *ui = *p++;
                 ui++;
@@ -654,22 +678,23 @@ int Parameters::testFileFont() const
         }
         else if (codesize == 1)
         {
-            unsigned char *pText8 = (unsigned char *)malloc(textSrc.getLength() * 4);
+            unsigned char *pText8 = (unsigned char *)malloc((textSrc.getLength() + 1) * 4);
             gr2::utf8::iterator ui = pText8;
             unsigned int *p = pText32;
-            for (int i = 0; i < textSrc.getLength(); ++i)
+            for (unsigned int i = 0; i < textSrc.getLength(); ++i)
             {
                 *ui = *p++;
                 ui++;
             }
             *ui = 0;
             pSeg = gr_make_seg(sizedFont, face, 0, features ? featureList : NULL, (gr_encform)codesize, pText8, textSrc.getLength(), rtl ? 1 : 0);
+            free(pText8);
         }
         else
             pSeg = gr_make_seg(sizedFont, face, 0, features ? featureList : NULL, textSrc.utfEncodingForm(),
                 textSrc.get_utf_buffer_begin(), textSrc.getLength(), rtl ? 1 : 0);
 
-        if (pSeg)
+        if (pSeg && !noprint)
         {
             int i = 0;
             float advanceWidth;
@@ -695,13 +720,15 @@ int Parameters::testFileFont() const
                 assert((i + 1 < numSlots) || (slot == gr_seg_last_slot(pSeg)));
                 float orgX = gr_slot_origin_X(slot);
                 float orgY = gr_slot_origin_Y(slot);
+                const gr_char_info *cinfo = gr_seg_cinfo(pSeg, gr_slot_original(slot));
                 fprintf(log, "%02d  %4d %3d@%d,%d\t%6.1f\t%6.1f\t%2d%4d\t%3d %3d\t",
                         i, gr_slot_gid(slot), lookup(map, (size_t)gr_slot_attached_to(slot)),
                         gr_slot_attr(slot, pSeg, gr_slatAttX, 0),
                         gr_slot_attr(slot, pSeg, gr_slatAttY, 0), orgX, orgY, gr_slot_can_insert_before(slot) ? 1 : 0,
-                        gr_cinfo_break_weight(gr_seg_cinfo(pSeg, gr_slot_original(slot))), gr_slot_before(slot), gr_slot_after(slot));
-               
-                if (pText32 != NULL)
+                        cinfo ? gr_cinfo_break_weight(cinfo) : 0, gr_slot_before(slot), gr_slot_after(slot));
+
+                if (pText32 != NULL && gr_slot_before(slot) + offset < charLength
+                                    && gr_slot_after(slot) + offset < charLength)
                 {
                     fprintf(log, "%7x\t%7x",
                         pText32[gr_slot_before(slot) + offset],
@@ -714,15 +741,16 @@ int Parameters::testFileFont() const
             // position arrays must be one bigger than what countGlyphs() returned
             fprintf(log, "Advance width = %6.1f\n", advanceWidth);
             unsigned int numchar = gr_seg_n_cinfo(pSeg);
-            fprintf(log, "\nChar\tUnicode\tBefore\tAfter\n");
+            fprintf(log, "\nChar\tUnicode\tBefore\tAfter\tBase\n");
             for (unsigned int j = 0; j < numchar; j++)
             {
                 const gr_char_info *c = gr_seg_cinfo(pSeg, j);
-                fprintf(log, "%d\t%04X\t%d\t%d\n", j, gr_cinfo_unicode_char(c), gr_cinfo_before(c), gr_cinfo_after(c));
+                fprintf(log, "%d\t%04X\t%d\t%d\t%d\n", j, gr_cinfo_unicode_char(c), gr_cinfo_before(c), gr_cinfo_after(c), int(gr_cinfo_base(c)));
             }
             free(map);
-            gr_seg_destroy(pSeg);
         }
+        if (pSeg)
+            gr_seg_destroy(pSeg);
         if (featureList) gr_featureval_destroy(featureList);
         gr_font_destroy(sizedFont);
         if (trace) gr_stop_logging(face);
@@ -741,9 +769,9 @@ int lookup(size_t *map, size_t val)
 
 int main(int argc, char *argv[])
 {
-    
+
     Parameters parameters;
-    
+
     if (!parameters.loadFromArgs(argc, argv))
     {
         fprintf(stderr,"Usage: %s [options] fontfile utf8text \n",argv[0]);
@@ -752,6 +780,8 @@ int main(int argc, char *argv[])
         fprintf(stderr,"-pt d\tPoint size (12)\n");
         fprintf(stderr,"-codes\tEnter text as hex code points instead of utf8 (false)\n");
         fprintf(stderr,"\te.g. %s font.ttf -codes 1000 102f\n",argv[0]);
+        fprintf(stderr,"-auto\tAutomatically generate a test string of all codes 1-0xFFF\n");
+        fprintf(stderr,"-noprint\tDon't print results\n");
         //fprintf(stderr,"-ls\tStart of line = true (false)\n");
         //fprintf(stderr,"-le\tEnd of line = true (false)\n");
         fprintf(stderr,"-rtl\tRight to left = true (false)\n");
@@ -761,12 +791,10 @@ int main(int argc, char *argv[])
         fprintf(stderr,"\nIf a font, but no text is specified, then a list of features will be shown.\n");
         fprintf(stderr,"-feat f=g\tSet feature f to value g. Separate multiple features with ,\n");
         fprintf(stderr,"-log out.log\tSet log file to use rather than stdout\n");
-        fprintf(stderr,"-trace trace.xml\tDefine a file for the XML trace log\n");
+        fprintf(stderr,"-trace trace.json\tDefine a file for the JSON trace log\n");
         fprintf(stderr,"-demand\tDemand load glyphs and cmap cache\n");
-        fprintf(stderr,"-cache\tEnable Segment Cache\n");
         fprintf(stderr,"-bytes\tword size for character transfer [1,2,4] defaults to 4\n");
         return 1;
     }
     return parameters.testFileFont();
 }
-
